package com.streambase.contrib.adapters.kxsystems;

import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;

import kx.c;
import kx.c.KException;

import com.streambase.sb.CompleteDataType;
import com.streambase.sb.DataType;
import com.streambase.sb.NullValueException;
import com.streambase.sb.Schema;
import com.streambase.sb.Timestamp;
import com.streambase.sb.Tuple;
import com.streambase.sb.TupleException;
import com.streambase.sb.Schema.Field;
import com.streambase.sb.operator.TypecheckException;

/**
 * The <code>QdbcHelper</code> class provides utility methods useful for integrating
 * Kdb+ and Q via Qdbc with StreamBase Adapters.
 * 
 * @author Darach Ennis (darach.ennis@streambase.com)
 *
 */
public class QdbcHelper {

	private QdbcHelper() {;}

	/**
	 * The <code>setFieldValue</code> static helper method maps a 'tuple field' returned from a Qdbc
	 * query result set into the corresponding field in a StreamBase tuple. The responsibility of matching
	 * the Qdbc result set field with the appropriate value lies with the caller and is assumed to be
	 * valid by this method. All StreamBase types except blob, function, capture, tuple, and list are
	 * supported. Only scalar / primitive Qdbc types are supported.
	 * 
	 * @param pivot 				The reusable (cached) StreamBase tuple 
	 * @param field				The field whose value is to be set
	 * @param objectArray			The Qdbc 'tuple field array' where the value is an element
	 * @param index				The index into the Qdbc 'tuple field array' where the value can be located
	 * @throws NullValueException
	 * @throws TupleException	
	 * @throws TypecheckException
	 */
	public static void setFieldValue(Tuple pivot, Field field, Object objectArray, int index) 
			throws NullValueException, TupleException, TypecheckException {
		
		Class<?> qClass = objectArray.getClass().getComponentType();
		
		if (Boolean.class.equals(qClass)) {
			pivot.setBoolean(field, ((Boolean[])objectArray)[index]);
			return;
		}
			
		if (boolean.class.equals(qClass)) {
			pivot.setBoolean(field, ((boolean[])objectArray)[index]);
			return;
		}
	
		if (Integer.class.equals(qClass)) {
			pivot.setInt(field, ((Integer[])objectArray)[index]);
			return;
		}
			
		if (int.class.equals(qClass)) {
			pivot.setInt(field, ((int[])objectArray)[index]);
			return;
		}
	
		if (Long.class.equals(qClass)) {
			pivot.setLong(field, ((Long[])objectArray)[index]);
			return;
		}
			
		if (long.class.equals(qClass)) {
			pivot.setLong(field, ((long[])objectArray)[index]);
			return;
		}
	
		if (Double.class.equals(qClass)) {
			pivot.setDouble(field, ((Double[])objectArray)[index]);
			return;
		}
			
		if (double.class.equals(qClass)) {
			pivot.setDouble(field, ((double[])objectArray)[index]);
			return;
		}
	
		if (String.class.equals(qClass)) {
			pivot.setString(field, ((String[])objectArray)[index]);
			return;
		}

		if (java.sql.Timestamp.class.equals(qClass)) {
			CompleteDataType sbType = field.getCompleteDataType();
			if (sbType.equalTypes(CompleteDataType.forTimestamp())) {
				// truncates fractional seconds portion
				final Date timestamp = ((java.sql.Timestamp[])objectArray)[index];
				pivot.setTimestamp(field, new Timestamp(timestamp));
				return;
			} else if (sbType.equalTypes(CompleteDataType.forInt())) {
				final int nanos = (((java.sql.Timestamp[])objectArray)[index]).getNanos();
				pivot.setInt(field, nanos);
				return;
			} else if (field.getDataType().equals(DataType.TUPLE)) {
				// set the timestamp and nano values into the first timestamp and int fields
				// in the sub-tuple schema, respectively
				final int nanos = (((java.sql.Timestamp[])objectArray)[index]).getNanos();

				final java.sql.Timestamp timestamp = ((java.sql.Timestamp[])objectArray)[index];
				timestamp.setNanos(0);
				final Date date = (java.util.Date) timestamp;

				DataType[] types = field.getSchema().getFieldTypes();
				int tspos = indexof(types, DataType.TIMESTAMP);
				int intpos = indexof(types, DataType.INT);
				Tuple tsTuple = field.getSchema().createTuple();
				if (tspos >= 0) {
					tsTuple.setField(tspos, date);
				}
				if (intpos >= 0) {
					tsTuple.setField(intpos, nanos);
				}
				pivot.setTuple(field, tsTuple);
				return;
			}
		}
		
		if (java.sql.Time.class.equals(qClass)) {
			final java.sql.Time sqlTime = ((java.sql.Time[])objectArray)[index];
			pivot.setTimestamp(field, new Timestamp(sqlTime));
			return;
		}
		
		// Fallthrough
		throw new TypecheckException("Unable to map class " + qClass.getCanonicalName() + " to a StreamBase DataType");
	}

	private static SimpleDateFormat qTimestampFormat = new SimpleDateFormat("yyyy.MM.dd HH:mm:ss.SSS");
	private static SimpleDateFormat qTimestampFormatNoMillis = new SimpleDateFormat("yyyy.MM.dd HH:mm:ss");
	private static SimpleDateFormat qTimeFormat = new SimpleDateFormat("HH:mm:ss.SSS");
	
	public static Object getFieldValue(Class<?> ctype, Tuple tuple, Field field) throws NullValueException, TupleException, TypecheckException {
		DataType type = field.getDataType();
		switch(type) {
		case BLOB: throw new TypecheckException("Unsupported type BLOB");
		case BOOL: return tuple.getBoolean(field.getIndex());
		case INT: return tuple.getInt(field.getIndex());
		case LONG: return tuple.getLong(field.getIndex());
		case DOUBLE: return tuple.getDouble(field.getIndex());
		case STRING: return "`" + tuple.getString(field.getIndex()); // @NOTE: Requires translation w.r.t. Q semantics for non-trivial 'strings'
		case TIMESTAMP: {
			if (java.sql.Time.class.equals(ctype)) {
				Timestamp ts = (tuple.getTimestamp(field.getIndex()));
				java.sql.Timestamp sqlTs = ts.toSQLTimestamp();
				return qTimeFormat.format(sqlTs);
			} else if (java.sql.Timestamp.class.equals(ctype)) {
				return qTimestampFormat.format(tuple.getTimestamp(field.getIndex()).toDate()).replace(" ", "D");
			} else {
				throw new TypecheckException("Unsupported Q Java type: " + ctype);
			}
		}
		case TUPLE: {
			DataType[] types = field.getSchema().getFieldTypes();
			int tspos = indexof(types, DataType.TIMESTAMP);
			int intpos = indexof(types, DataType.INT);
			String ts = null;
			if (tspos >= 0) {
				try {
					ts = qTimestampFormatNoMillis.format(tuple.getTuple(field.getIndex()).getTimestamp(tspos).toDate()).replace(" ", "D");
				} catch (NullValueException nve) {
					ts = "0";
				}
			} else {
				ts = "0";
			}
			if (intpos > 0) {
				try {
					ts += "." + String.format(Locale.US, "%09d", tuple.getTuple(field.getIndex()).getInt(intpos));
					// Locale.US only to guarantee we have Arabic numerals
				} catch(NullValueException nve) {
					// leave well enough alone
				}
			}
			return ts;
		}
		default: throw new TypecheckException("Unsupported StreamBase type " + type);
		}
	}

	/**
	 * The <code>deduceSchema</code> method, given a Qdbc connection, a StreamBase schema name
	 * and a Qdbc query result set introspects and generates the corresponding StreamBase representation.
	 * 
	 * Only flat scalar field values are supported at this time. All StreamBase scalar types are supported
	 * with the exception of Blob. Hierarchical StreamBase structures are not supported at this time.
	 * 
	 * Q tables are implictly typed when declared. The first insertion sets and seals type information for
	 * all proceeding tuple mutations. 
	 * 
	 * @param c					The Q Qdbc connection
	 * @param schemaName			The StreamBase schema name for the schema being generated
	 * @param qResultSet			The (empty) Qdbc 'meta' query result set with meta type information that seeds this translation
	 * @return						The corresponding StreamBase schema representation
	 * @throws IOException 
	 * @throws KException 
	 * @throws TypecheckException 
	 * @throws TypecheckException
	 * @throws KException
	 * @throws IOException
	 * @throws TupleException 
	 */
	public static Schema deduceSchema(c c, String schemaName, c.Flip qResultSet) throws TypecheckException, KException, IOException, TupleException {
		return deduceSchema(c, schemaName, qResultSet, null);
	}

	/**
	 * The <code>deduceSchema</code> method, given a Qdbc connection, a StreamBase schema name
	 * and a Qdbc query result set introspects and generates the corresponding StreamBase representation.
	 * 
	 * Only flat scalar field values are supported at this time. All StreamBase scalar types are supported
	 * with the exception of Blob. Hierarchical StreamBase structures are not supported at this time.
	 * 
	 * Q tables are implicitly typed when declared. The first insertion sets and seals type information for
	 * all proceeding tuple mutations.
	 * 
	 * If the refSchema is null or has no fields, then the generated Schema has the names of the fields from
	 * the qResultSet, and the data types have a default type based on the Q type of the qResultSet field.
	 * 
	 * If the refSchema has fields, then the generated schema is a copy of refSchema.
	 * 
	 * @param c					The Q Qdbc connection
	 * @param schemaName		The StreamBase schema name for the schema being generated
	 * @param qResultSet		The (empty) Qdbc 'meta' query result set with meta type information that seeds this translation
	 * @param refSchema			An optional schema reference to be used for specifying name and type of returned schema fields,
	 * 							overriding the name and type of the qResultSet fields in the corresponding position 
	 * @return					The corresponding StreamBase schema representation
	 * @throws TypecheckException
	 * @throws KException
	 * @throws IOException
	 * @throws TupleException 
	 */
	public static Schema deduceSchema(c c, String schemaName, c.Flip qResultSet, Schema refSchema)
				throws TypecheckException, KException, IOException, TupleException {
		if ((refSchema != null) && (refSchema.getFieldCount() == 0)) {
			// override a zero-length schema with Q result set schema
			refSchema = null;
		}
		final int numQFields = qResultSet.x.length;
		if (refSchema != null && refSchema.getFieldCount() > numQFields) {
			throw new TypecheckException("User defined schema cannot have more fields than Qdbc source schema.");
		}
		Field[] fields = new Field[(refSchema == null) ? numQFields : refSchema.getFieldCount()];
		int i = 0;
		for (int j = 0; j < qResultSet.y.length; j++) {
			if (refSchema != null && !refSchema.hasField(qResultSet.x[j])) {
				continue;
			}
			if (qResultSet.y[j].getClass().isArray()) {
				Class<?> ctype = qResultSet.y[j].getClass().getComponentType();
				fields[i] = Schema.createField(QdbcHelper.translateDatatype(ctype), qResultSet.x[j]);
				if (refSchema != null) {
					Schema.Field refSchemaField = refSchema.getField(qResultSet.x[j]);
					// only support sub-tuples for Q timestamp translation, for now
					if ((DataType.TUPLE.equals(refSchemaField.getDataType())) && java.sql.Timestamp.class.equals(ctype)) {
						fields[i] = Schema.createTupleField(fields[i].getName(), refSchemaField.getSchema());
					} else {
						fields[i] = Schema.createField(fields[i].getDataType(), fields[i].getName());
					}
				}
			} else {
				throw new TypecheckException("Invalid or unsupported Qdbc result structure.");
			}
			i++;
		}
		return new Schema(schemaName, fields);
	}

	/**
	 * The <code>translateDatatype</code> method given the Qdbc component type of a Q Qdbc
	 * query result set field returns the corresponding StreamBase DataType.
	 * 
	 * @param fromQ				The Q Qdbc query result set field component type to be translated
	 * @return						The corresponding StreamBase data type
	 * @throws TypecheckException
	 */
	public static DataType translateDatatype(Class<?> fromQ) throws TypecheckException {
		// 'BLOB' is not supported at this time
		// embedded Lists or Dicts (hierarchical structures) are not supported
		
		if (Boolean.class.equals(fromQ) || boolean.class.equals(fromQ)) {
			return DataType.BOOL;
		}
		
		if (Integer.class.equals(fromQ) || int.class.equals(fromQ)) {
			return DataType.INT;
		}

		if (Long.class.equals(fromQ) || long.class.equals(fromQ)) {
			return DataType.LONG;
		}

		if (Double.class.equals(fromQ) || double.class.equals(fromQ)) {
			return DataType.DOUBLE;
		}

		if (String.class.equals(fromQ)) {
			return DataType.STRING;
		}

		if (java.sql.Time.class.equals(fromQ)) {
			return DataType.TIMESTAMP;
		}
		
		if (java.sql.Timestamp.class.equals(fromQ)) {
			return DataType.TIMESTAMP;
		}
		
		// Fallthrough
		throw new TypecheckException("Unable to map class " + fromQ.getCanonicalName() + " to a StreamBase DataType");
	}

	/**
	 * The <code>connect</code> method creates a new Qdbc connection. If hostname is the empty string
	 * then localhost is assumed. Otherwise; the specified host is used to attempt terminating the
	 * a Qdbc connection. The port on which Q is listening for Qdbc client connections.
	 * 
	 * Transport Level Security via TLS 1.0 / SSL 3.0 and authenticated connections are not
	 * supported at this time.
	 * 
	 * @param hostname				The hostname to terminate the Qdbc client connection against
	 * @param port					The TCP port over which the Qdbc service is provided
	 * @return						A Q Qdbc connection handle
	 * @throws KException
	 * @throws IOException
	 */
	public static c connect(String hostname, int port) throws KException, IOException {
		return new c(hostname, port);
	}

	/**
	 * The <code>connect</code> method creates a new Qdbc connection.
	 * If username is null or empty and useTLS is false, then connect
	 * using only the hostname and port (which tells the API to use the
	 * user.name system property as a source of username and password)
	 * 
	 * @param hostname				The hostname to terminate the Qdbc client connection against
	 * @param port					The TCP port over which the Qdbc service is provided
	 * @param username				The username to use to authenticate the the kdb+ server
	 * @param password				The password to use to authenticate the the kdb+ server
	 * @param useTLS				Whether to use TLS/SSL for the connection to the kdb+ server
	 * @return						A Q Qdbc connection handle
	 * @throws KException
	 * @throws IOException
	 */
	public static c connect(String hostname, int port, String username, String password, boolean useTLS) throws KException, IOException {
		if (((username == null) || (username.length() == 0)) && !useTLS) {
			return new c(hostname, port);
		} else {
			return new c(hostname,
					 port,
					 username + (((password != null) && (password.length() > 0)) ? ":"+password : ""),
					 useTLS
				    );
		}
	}

	
	/**
	 * The <code>toDelimitedString</code> method serializes tuple field values separated by a specified
	 * character delimiter.
	 * 
	 * @param tuple				The tuple for which fields require stringification
	 * @param sep					The character separator to be inserted between stringified fields
	 * @return						The resultant string
	 * @throws TupleException
	 * @throws TypecheckException
	 */
	public static String toDelimitedString(c.Flip qResultSet, Tuple tuple, char sep) throws TupleException, TypecheckException {
		return toDelimitedString(qResultSet, tuple, ""+ sep);
	}

	/**
	 * The <code>toDelimitedString</code> method serializes tuple field values separated by a specified
	 * string delimiter.
	 * 
	 * @param tuple				The tuple for which fields require stringification
	 * @param sep					The string separator to be inserted between stringified fields
	 * @return						The resultant string
	 * @throws TupleException
	 * @throws TypecheckException
	 */
	public static String toDelimitedString(c.Flip qResultSet, Tuple tuple, String sep) throws TupleException, TypecheckException {
		StringBuilder sb = new StringBuilder();
		int j = 0;
		for (Field field : tuple.getSchema().getFields()) {
			sb.append(getFieldValue(qResultSet.y[j].getClass().getComponentType(), tuple, field));
			sb.append(sep);
			j++;
		}
		sb.replace(sb.length()-1, sb.length(), "");
		return sb.toString();
	}

	/**
	 * The <code>toDelimitedString</code> method serializes tuple field name and value pairs separated by a specified
	 * string name/value delimiter and a field separator.
	 * 
	 * @param tuple				The tuple for which fields require stringification
	 * @param sep					The character separator to be inserted between stringified fields
	 * @return						The resultant string
	 * @throws TupleException
	 * @throws TypecheckException
	 */
	public static String toDelimitedString(c.Flip qResultSet, Tuple tuple, String delim,
			String sep) throws TupleException, TypecheckException {
		StringBuilder sb = new StringBuilder();
		int j = 0;
		for (Field field : tuple.getSchema().getFields()) {
			sb.append(field.getName());
			sb.append(delim);
			sb.append(getFieldValue(qResultSet.y[j].getClass().getComponentType(), tuple, field));
			sb.append(sep);
			j++;
		}
		sb.replace(sb.length()-1, sb.length(), sep);
		return sb.toString();
	}
	
	private static int indexof(DataType[] dts, DataType dt) {
		if (dts == null )  {
			throw new NullPointerException("Field type array is null");
		}
		if (dt == null )  {
			throw new NullPointerException("data type is null");
		}
		
		for (int i = 0; i < dts.length; i++) {
			if (dts[i] == dt) {
				return i;
			}
		}
		return -1;
	}

}
