package com.streambase.contrib.adapters.kxsystems;

import java.io.IOException;
import java.lang.reflect.Array;

import kx.c;
import kx.c.KException;


import com.streambase.sb.*;
import com.streambase.sb.Schema.Field;
import com.streambase.sb.operator.*;

import static com.streambase.contrib.adapters.kxsystems.QdbcConstants.*;

/**
 * '@Note' meta-annotation comments are employed where noteworthy choices have been
 * taken and explanations are worth sharing.
 * 
 * @author Darach Ennis (darach.ennis@streambase.com)
 */
public class QdbcReader extends QdbcReaderBase {

	public static final long serialVersionUID = 1208789799402L;

	private String displayName = "Qdbc Reader";
	
	private String metaQuery;
	private String queryTable;
	private String Mode;
	private long pollingMillis = _POLLING_MILLIS_ONCE;

	// Constants for enum property conversionType
	public static String _ENUM_PROP_BLOCKING_ = "BLOCKING";
	public static String _ENUM_PROP_STREAMING_ = "STREAMING"; // @NOTE: Not supported at this time

	public static long _POLLING_MILLIS_ONCE = 0L;
	
	private c.Flip qdbcMeta = null;
	
	/**
	 */
	public QdbcReader() {
		super();
		setPortHints(0, 1);
		setDisplayName(displayName);
		setHostname("localhost");
		setMetaQuery("");
		setQueryTable("");
		setPort(DEFAULT_QDBC_PORT);
		setMode(_ENUM_PROP_BLOCKING_);
		setPollingMillis(_POLLING_MILLIS_ONCE);
		setLogLevel(LogLevel.INFO);
	}

	/**
	 * @throws IOException 
	 * @throws KException 
	 */
	public void typecheck() throws TypecheckException {		
		if (!_ENUM_PROP_BLOCKING_.equals(getMode())) {
			throw new TypecheckException("Qdbc does not support streaming at this time");
		}

		conn = checkAndConnect();

		if (queryTable == null || queryTable.trim().equalsIgnoreCase("")) {
			throw new TypecheckException("Qdbc query must be a valid Q query");
		}
	
		if (qdbcMeta == null) { // cache q meta data
			try {
				qdbcMeta = (c.Flip)conn.k(getMetaQuery()); // @NOTE - Intent is to use a lightweight metadata providing only query such as 'select from <table> where 1 == 0'
			} catch (KException e) {
				logger.error("Unable to connect to Q host at " + getHostname() + ":" + getPort(), e);
				throw new TypecheckException("Qdbc query must be a valid Q query", e);
			} catch (IOException e) {
				logger.error("Unable to connect to Q host at " + getHostname() + ":" + getPort(), e);
				throw new TypecheckException("Qdbc query must be a valid Q query", e);
			} 
			
			try {
					schema0 = QdbcHelper.deduceSchema(conn, getName()+":0", qdbcMeta, getSchema0());
			} catch (TupleException e) {
				logger.error("Unable to deduce schema from Qdbc metadata query", e);
				throw new TypecheckException("Qdbc query must use StreamBase compatible types.", e);
			} catch (KException e) {
				logger.error("Unable to deduce schema from Qdbc metadata query", e);
				throw new TypecheckException("Qdbc query must use StreamBase compatible types.", e);
			} catch (IOException e) {
				logger.error("Unable to deduce schema from Qdbc metadata query", e);
				throw new TypecheckException("Qdbc query must use StreamBase compatible types.", e);
			}
		}
		
		// Sets the output schema for the given output port (port #'s are zero based).
		setOutputSchema(0, schema0);
		
	}

	/**
	 */
	public void run() {
		fillAndOutputTuple();
	}

	protected void fillAndOutputTuple() {
		Tuple pivot = null;
		pivot = getRuntimeOutputSchema(0).createTuple();
		boolean first = true;

		try {
			while (shouldRun()) {
				if (first || getPollingMillis() > 0L) {
					first = false;
					c.Flip qdbcStatement;
					try {
						qdbcStatement = (c.Flip)conn.k(getQueryTable());
					} catch (KException e) {
						logger.error("Malformed Q query attempted with query: " + getQueryTable() + e);
						throw new StreamBaseRuntimeException("Qdbc query must be a valid Q query", e);
					} catch (IOException e) {
						logger.error("Malformed Q query attempted with query: " + getQueryTable() + e);
						throw new StreamBaseRuntimeException("Qdbc query must be a valid Q query", e);
					} 
					try {
						int count = 0;
						if (qdbcStatement.y[0].getClass().isArray()) {
							count = Array.getLength(qdbcStatement.y[0]);
						} else {
							throw new TypecheckException("Invalid or unsupported Qdbc result structure.");
						}

						int j = 0;
						for (; j < count; j++) {
							pivot.clear();
							String[] fieldNames = (String[])qdbcStatement.x;
							for (int k = 0; k < fieldNames.length; k++) {
								if (getSchema0().hasField(fieldNames[k])) {
									final Field field = getSchema0().getField(fieldNames[k]);
									QdbcHelper.setFieldValue(pivot, field, qdbcStatement.y[k], j);
								}
							}
							if (logger.isDebugEnabled())
								logger.debug(getName()+ ": sending tuple " + pivot);
							sendOutput(0, pivot);
						}
					} catch (StreamBaseException e) {
						logger.error("Exception sending output to port " + 0, e);
						return;
					}
				}

				try {
					Thread.sleep((getPollingMillis() == _POLLING_MILLIS_ONCE) ? 1000L : getPollingMillis()); 
				} catch (InterruptedException e) {
				}
			}
		} finally {
			try {
				conn.close();
			} catch (Exception ignoreAndContinue) {;}
		}
	}

	public void setMetaQuery(String metaQuery) {
		this.metaQuery = metaQuery;
	}
	
	public String getMetaQuery() {
		return this.metaQuery;
	}

	public void setQueryTable(String queryTable) {
		this.queryTable = queryTable;
	}
	
	public String getQueryTable() {
		return this.queryTable;
	}
	
	public void setMode(String Mode) {
		this.Mode = Mode;
	}

	public String getMode() {
		return this.Mode;
	}

	public long getPollingMillis() {
		return pollingMillis;
	}

	public void setPollingMillis(long pollingMillis) {
		this.pollingMillis = pollingMillis;
	}
}
