package com.streambase.contrib.adapters.kxsystems;

import static com.streambase.contrib.adapters.kxsystems.QdbcConstants.MAX_TCP_PORT;
import static com.streambase.contrib.adapters.kxsystems.QdbcConstants.MAX_WELLKNOWN_TCP_PORT;
import static com.streambase.contrib.adapters.kxsystems.QdbcConstants.MIN_TCP_PORT;

import java.io.IOException;

import org.slf4j.Logger;

import kx.c;
import kx.c.KException;

import com.streambase.sb.Schema;
import com.streambase.sb.StreamBaseException;
import com.streambase.sb.StreamBaseRuntimeException;
import com.streambase.sb.adapter.InputAdapter;
import com.streambase.sb.operator.Parameterizable;
import com.streambase.sb.operator.TypecheckException;

/**
 * The <code>QdbcReaderBase</code> is a base class for all QdbcReaders.
 * 
 * @author Darach Ennis (darach.ennis@streambase.com)
 *
 */
public abstract class QdbcReaderBase extends InputAdapter implements Parameterizable, Runnable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 2022172677507607578L;
	
	protected Logger logger;
	private String hostname;
	private int port;
	private String username;
	private String password;
	private boolean useTLS;
    private LogLevel logLevel = LogLevel.INFO;
	protected Schema schema0;

	protected c conn = null;

	public QdbcReaderBase() {
		super();
		logger = getLogger();
	}

	protected c checkAndConnect() throws TypecheckException {
		if (getPort() < MIN_TCP_PORT || getPort() > MAX_TCP_PORT) {
			logger.error("Attempt to connect to an invalid port range");
		}
		
		if (getPort() <= MAX_WELLKNOWN_TCP_PORT) {
			logger.warn("The well known port, " + port + ", may be required by another service.");
		}
		
		try {
			return QdbcHelper.connect(getHostname(), getPort(), getUsername(), getPassword(), isUseTLS());
		} catch (KException e) {
			logger.error("Unable to connect to Q host at " + getHostname() + ":" + getPort(), e);
			throw new TypecheckException("Cannot connect via Qdbc to " + getHostname() + ":"  + getPort(), e);
		} catch (IOException e) {
			logger.error("Unable to connect to Q host at " + getHostname() + ":" + getPort(), e);
			throw new TypecheckException("Cannot connect via Qdbc to " + getHostname() + ":"  + getPort(), e);
		}
	}

	/**
	 * Initialize the adapter. If typecheck succeeds, the init method is called before
	 * the StreamBase application is started. Note that your adapter is not required to
	 * define the init method, unless you need to register a runnable or perform
	 * initialization of a resource such as, for example, a JDBC pool.
	 */
	public void init() throws StreamBaseException {
		super.init();
		
		try {
			setLogLevel(logger, getLogLevel());
		} catch (StreamBaseException e) {
		            logger.warn(String.format("%s: Failed to set the log level to %s, most likely because a logging backend other than the default is being used. Current StreamBase log level will be used instead.",
		            		    getName(),
		            		    getLogLevel()));
	    }			
	
		// Register the object so it will be run as a thread managed by StreamBase.
		registerRunnable(this, true);
	}

	/**
	 * Shutdown adapter by cleaning up any resources used (e.g. close an output file
	 * if it has been opened). When the application is shutting down, the adapter's shutdown.
	 * method will be called first. Once this has returned, all threads should exit and the
	 * adapter is considered shutdown.
	 */
	public void shutdown() {
		try {
			conn.close();
		} catch (IOException ignoreAndContinue) {
			logger.error("Failed to close Q connection during shutdown phase, ignoring to continue shutdown", ignoreAndContinue);
		}
	}

	/**
	 * This method will be called when an adapter starts or resumes execution,
	 * before any registered runnables are started or resumed.  Note that if the
	 * application is shutdown directly from a suspended state, this will
	 * <b>not</b> be called; instead {@link #shutdown()} will be called.
	 *
	 * resume() is a callback that is called by the StreamBase runtime.
	 */
	public void resume() {
		try {
			conn = QdbcHelper.connect(getHostname(),getPort(),getUsername(),getPassword(),isUseTLS());
		} catch (KException e) {
			throw new StreamBaseRuntimeException(e);
		} catch (IOException e) {
			throw new StreamBaseRuntimeException(e);
		}
	}

	/**
	 * This method will be called after all registered runnables of
	 * the adapter have resumed.  That is, once {@link #shouldRun()}
	 * has unblocked and returned true in all registered runnables.
	 *
	 * resumed() is a callback that is called by the StreamBase runtime.
	 */
	public void resumed() {
		// Explicitly left blank
	}

	/**
	 * This method will be called when an adapter suspends, before any
	 * registered runnables are suspended.
	 *
	 * suspend() is a callback that is called by the StreamBase runtime.
	 */
	public void suspend() {
		try {
			conn.close();
		} catch (IOException e) {
			throw new StreamBaseRuntimeException(e);
		}
	}

	/**
	 * suspended() will be called after all registered runnables of
	 * the adapter have suspended.  That is, once all registered
	 * runnables have called and are blocked in {@link #shouldRun()}.
	 *
	 * suspended() is a callback that is called by the StreamBase runtime.
	 */
	public void suspended() {
		// Explicitly left blank
	}

	/***************************************************************************************
	 * The getter and setter methods provided by the Parameterizable object.               *
	 * StreamBase Studio uses them to determine the name and type of each property         *
	 * and obviously, to set and get the property values.                                  *
	 ***************************************************************************************/
	public void setHostname(String hostname) {
		this.hostname = hostname;
	}

	public String getHostname() {
		return this.hostname;
	}

	public void setPort(int port) {
		this.port = port;
	}

	public int getPort() {
		return this.port;
	}

	public String getUsername() {
		return username;
	}

	public void setUsername(String username) {
		this.username = username;
	}

	public String getPassword() {
		return password;
	}

	public void setPassword(String password) {
		this.password = password;
	}

	public boolean isUseTLS() {
		return useTLS;
	}

	public void setUseTLS(boolean useTLS) {
		this.useTLS = useTLS;
	}

	public LogLevel getLogLevel() {
	    return this.logLevel;
	}
	    
	public void setLogLevel(LogLevel level) {
	    this.logLevel = level;
	}
	 
	public Schema getSchema0() {
		return this.schema0;
	}

	public void setSchema0(Schema schema0) {
		this.schema0 = schema0;
	}

}