package com.streambase.contrib.adapters.kxsystems;

import java.io.IOException;
import java.lang.reflect.Array;

import com.streambase.sb.NullValueException;
import com.streambase.sb.Schema;
import com.streambase.sb.Tuple;
import com.streambase.sb.TupleException;
import com.streambase.sb.Schema.Field;
import com.streambase.sb.operator.TypecheckException;

import kx.c;
import kx.c.KException;

/**
 * The <code>QdbcTest</code> class was used to test drive the <code>QdbcHelper</code> class
 * than translates from the native Qdbc. Qdbc is an efficient database connectivity driver for
 * the Q programming language and Kdb+ database.
 * 
 * The Q script used to drive the Qdbc utility code was as follows:
 * 
 * <pre>
 * / generate a table of 1M randomly generated tuples
 * n:1000000
 * stocks:`goog`ibm`msft`ebay`amzn`intel`amd`iona`jpmc
 * trade:([]symbol:n?stocks; price:n?100.0;amount:100*10+n?20;time:n?24:00:00.000)
 * </pre>
 * 
 * This program creates a 'stock' in memory table and generates a million randomly generated
 * trades and inserts them into the database. The table can be exposed to Jdbc,Odbc,C,Java,.NET
 * and HTTP based hosts supporting Csv,Xml,Html by running the following commands within a
 * shell environment:
 * 
 * <pre>
 * dennis@debdev:~/qtrain$ q -p 5001
 * KDB+ 2.4 2008.03.31 Copyright (C) 1993-2008 Kx Systems
 * l32/ 1()core 504MB monkey debdev 127.0.0.1 TIMEOUT 2009.04.01
 * 
 * q)\l gen.q / load the gen.q script
 * q)trade
 * symbol price    amount time
 * -----------------------------------
 * amzn   82.64632 1500   02:28:10.187
 * jpmc   50.47438 1600   00:30:28.079
 * msft   83.78846 2100   18:41:52.225
 * ebay   69.91293 2800   10:12:32.620
 * iona   50.60001 1300   17:56:32.126
 * ebay   58.65375 2100   06:15:10.504
 * q)
 * </pre>
 * 
 * Q will start server processes listening automagically as a port (-p 5001) has been specified
 * on the command line to Q. The Q environment can be browsed via pointing a web browser at
 * <code>http://localhost:5001/</code>
 * 
 * 
 * @author Darach Ennis (darach.ennis@streambase.com)
 *
 */
public class QdbcTest {
	public static void main(String args[]) throws KException, IOException, TypecheckException, TupleException {
		c c = null;
		try {
			// properties
			c = QdbcHelper.connect("", 5001);

			long start = System.currentTimeMillis();
			
			c.Flip dol = (c.Flip)c.k("trade");
			Schema schema1 = QdbcHelper.deduceSchema(c, "schema1", dol);
			Tuple pivot1 = schema1.createTuple();
			int emitted = QdbcTest.emitTuples(dol, schema1, pivot1);
			System.out.println("Schema1: " + emitted);

			long stop = System.currentTimeMillis() - start;
			System.out.println("1M tuples in " + stop + "ms");

			start = System.currentTimeMillis();
			
			dol = (c.Flip)c.k("select from trade where symbol in (`ibm)"); // circa 10% of select(*) on average
			Schema schema2 = QdbcHelper.deduceSchema(c, "schema2", dol);
			Tuple pivot2 = schema2.createTuple();
			emitted = QdbcTest.emitTuples(dol, schema2, pivot2);
			System.out.println("Schema2: " + emitted);

			stop = System.currentTimeMillis() - start;
			System.out.println("~100K tuples in " + stop + "ms");
			
		} finally {
			c.close();
		}
	}

	/**
	 * The <code>emitTuples</code> method captures the required Qdbc logic that
	 * requires integration with the StreamBase Adapter.run() method.
	 * 
	 * @param dol
	 * @param schema1
	 * @param pivot
	 * @return
	 * @throws TypecheckException
	 * @throws TupleException
	 * @throws NullValueException
	 */
	public static int emitTuples(c.Flip dol, Schema schema1, Tuple pivot)
			throws TypecheckException, TupleException, NullValueException {
		int length = 0;
		if (dol.y[0].getClass().isArray()) {
			length = Array.getLength(dol.y[0]);
		} else {
			throw new TypecheckException("Invalid or unsupported Qdbc result structure.");
		}
		
		int j = 0;
		for (; j < length; j++) {
			pivot.clear();
			for (int i = 0; i < dol.x.length; i++) {
				final Field field = schema1.getField(i);
				QdbcHelper.setFieldValue(pivot, field, dol.y[i], j);
			}
		}
		return j;
	}
}
