package com.streambase.contrib.adapters.kxsystems;

import static com.streambase.contrib.adapters.kxsystems.QdbcConstants.MAX_TCP_PORT;
import static com.streambase.contrib.adapters.kxsystems.QdbcConstants.MAX_WELLKNOWN_TCP_PORT;
import static com.streambase.contrib.adapters.kxsystems.QdbcConstants.MIN_TCP_PORT;

import java.io.IOException;

import org.slf4j.Logger;

import com.streambase.sb.Schema;
import com.streambase.sb.StreamBaseException;
import com.streambase.sb.StreamBaseRuntimeException;
import com.streambase.sb.Tuple;
import com.streambase.sb.adapter.OutputAdapter;
import com.streambase.sb.operator.Parameterizable;
import com.streambase.sb.operator.TypecheckException;

import kx.c;
import kx.c.KException;

/**
 * The <code>QdbcWriterBase</code> is a base class for all QdbcWriters.
 * 
 * @author Darach Ennis (darach.ennis@streambase.com)
 *
 */
public abstract class QdbcWriterBase extends OutputAdapter implements Parameterizable {

	private static final long serialVersionUID = 6344702221522478752L;
	
	protected Logger logger;
	private String hostname;
	private int port;
	private String username;
	private String password;
	private boolean useTLS;
    private LogLevel logLevel = LogLevel.INFO;
	protected Schema schema0;
	
	protected c conn = null;
	protected Tuple pivot = null;

	public QdbcWriterBase() {
		super();
		logger = getLogger();
	}

	protected c checkAndConnect() throws TypecheckException {
		if (getPort() < MIN_TCP_PORT || getPort() > MAX_TCP_PORT) {
			logger.error("Attempt to connect to an invalid port range");
		}
		
		if (getPort() <= MAX_WELLKNOWN_TCP_PORT) {
			logger.warn("The well known port, " + port + ", may be required by another service.");
		}
		
		try {
			return QdbcHelper.connect(getHostname(), getPort(), getUsername(), getPassword(), isUseTLS());
		} catch (KException e) {
			logger.error("Unable to connect to Q host at " + getHostname() + ":" + getPort(), e);
			throw new TypecheckException("Cannot connect via Qdbc to " + getHostname() + ":"  + getPort(), e);
		} catch (IOException e) {
			logger.error("Unable to connect to Q host at " + getHostname() + ":" + getPort(), e);
			throw new TypecheckException("Cannot connect via Qdbc to " + getHostname() + ":"  + getPort(), e);
		}
	}

	/**
	 */
	public void init() throws StreamBaseException {
		super.init();
		try {
			setLogLevel(logger, getLogLevel());
		} catch (StreamBaseException e) {
		            logger.warn(String.format("%s: Failed to set the log level to %s, most likely because a logging backend other than the default is being used. Current StreamBase log level will be used instead.",
		            		    getName(),
		            		    getLogLevel()));
	    }			
		
	}

	/**
	 */
	public void shutdown() {
		try {
			conn.close();
		} catch (IOException ignoreAndContinue) {
			logger.error("Failed to close Q connection during shutdown phase, ignoring to continue shutdown", ignoreAndContinue);
		}
	}

	/**
	 */
	public void resume() {
		try {
			conn = QdbcHelper.connect(getHostname(),getPort(), getUsername(), getPassword(), isUseTLS());
		} catch (KException e) {
			throw new StreamBaseRuntimeException(e);
		} catch (IOException e) {
			throw new StreamBaseRuntimeException(e);
		}
	}

	public void resumed() {
		// Explicitly left blank
	}

	public void suspend() {
		try {
			conn.close();
		} catch (IOException e) {
			throw new StreamBaseRuntimeException(e);
		}
	}

	public void suspended() {
		// Explicitly left blank
	}

	/***************************************************************************************
	 * The getter and setter methods provided by the Parameterizable object.               *
	 * StreamBase Studio uses them to determine the name and type of each property         *
	 * and obviously, to set and get the property values.                                  *
	 ***************************************************************************************/
	public void setHostname(String hostname) {
		this.hostname = hostname;
	}

	public String getHostname() {
		return this.hostname;
	}

	public void setPort(int port) {
		this.port = port;
	}

	public int getPort() {
		return this.port;
	}

	public String getUsername() {
		return username;
	}

	public void setUsername(String username) {
		this.username = username;
	}

	public String getPassword() {
		return password;
	}

	public void setPassword(String password) {
		this.password = password;
	}

	public boolean isUseTLS() {
		return useTLS;
	}

	public void setUseTLS(boolean useTLS) {
		this.useTLS = useTLS;
	}

	public LogLevel getLogLevel() {
	    return this.logLevel;
	}
	    
	public void setLogLevel(LogLevel level) {
	    this.logLevel = level;
	}
	 
	public Schema getSchema0() {
		return this.schema0;
	}

	public void setSchema0(Schema schema0) {
		this.schema0 = schema0;
	}

}